from abstra_internals.interface.sdk.hooks import (
    get_raw_request,
    get_request,
    send_json,
    send_response,
)

__all__ = ["get_raw_request", "get_request", "send_response", "send_json"]



from typing import List, Optional

from abstra_internals.linter.linter import LinterFix, LinterIssue, LinterRule
from abstra_internals.repositories.requirements import RequirementsRepository


class MergeDuplicatePackages(LinterFix):
    pkg_version: Optional[str]
    pkg_name: str

    def __init__(self, pkg_name: str, pkg_version: Optional[str]) -> None:
        super().__init__()
        self.pkg_version = pkg_version
        self.pkg_name = pkg_name
        version_name = pkg_version if pkg_version is not None else "latest"
        self.label = f"Choose {version_name}"

    def fix(self):
        requirements = RequirementsRepository.load()
        requirements.ensure(self.pkg_name, self.pkg_version)
        requirements.delete_duplicates(self.pkg_name, self.pkg_version)
        RequirementsRepository.save(requirements)

    def __hash__(self) -> int:
        return hash((self.pkg_name, self.pkg_version))


class DuplicatePackagesInRequirementsFound(LinterIssue):
    type = "bug"

    def __init__(self, name: str, versions: List[Optional[str]]) -> None:
        self.label = f"Duplicate {name} found in requirements.txt"
        fixes = [
            MergeDuplicatePackages(pkg_name=name, pkg_version=version)
            for version in versions
        ]
        self.fixes = list(
            dict.fromkeys(fixes).keys()
        )  # remove duplicates keeping order


class DuplicatePackagesInRequirements(LinterRule):
    label = "Duplicate package in requirements.txt"
    type = "bug"

    def find_issues(self) -> List[LinterIssue]:
        requirements = RequirementsRepository.load()
        duplicates = requirements.get_duplicates()
        issues = []
        for name, versions in duplicates.items():
            issues.append(
                DuplicatePackagesInRequirementsFound(
                    name=name, versions=[r.version for r in versions]
                )
            )
        return issues



import typing
from datetime import date, datetime
from uuid import UUID

from abstra_internals.utils import serialize as json_serialize


@typing.runtime_checkable
class WithAsDict(typing.Protocol):
    @property
    def as_dict(self) -> dict: ...


def escape_ref(ref: str):
    return ref.replace('"', '""')


def quoted_identifier(ref: str):
    return f'"{escape_ref(ref)}"'


def serialize(value: typing.Any):
    if isinstance(value, WithAsDict):
        return serialize(value.as_dict)
    if isinstance(value, (datetime, date)):
        return value.isoformat()
    if isinstance(value, (dict, list, tuple)):
        return json_serialize(value)
    if isinstance(value, set):
        return json_serialize(list(value))
    if isinstance(value, UUID):
        return str(value)

    return value



from unittest import TestCase

from abstra_internals.repositories.requirements import Requirement, Requirements


class TestRequirementsRepository(TestCase):
    def test_parse_requirement(self):
        self.assertEqual(Requirement.from_text("foo"), Requirement(name="foo"))
        self.assertEqual(
            Requirement.from_text("foo==1.0.0"),
            Requirement(name="foo", version="1.0.0"),
        )

    def test_serialize_requirement(self):
        self.assertEqual(Requirement(name="foo").to_text(), "foo")
        self.assertEqual(
            Requirement(name="foo", version="1.0.0").to_text(), "foo==1.0.0"
        )

    def test_parse_requirements(self):
        self.assertEqual(
            Requirements.from_text("foo\nbar==1.0.0"),
            Requirements(
                libraries=[
                    Requirement(name="foo"),
                    Requirement(name="bar", version="1.0.0"),
                ]
            ),
        )

    def test_serialize_requirements(self):
        self.assertEqual(
            Requirements(
                libraries=[
                    Requirement(name="foo"),
                    Requirement(name="bar", version="1.0.0"),
                ]
            ).to_text(),
            "foo\nbar==1.0.0",
        )

    def test_parse_empty_requirements(self):
        self.assertEqual(Requirements.from_text(""), Requirements(libraries=[]))

    def test_parse_requirements_with_comments(self):
        self.assertEqual(
            Requirements.from_text("# foo\nfoo\n# bar\nbar==1.0.0"),
            Requirements(
                libraries=[
                    Requirement(name="foo"),
                    Requirement(name="bar", version="1.0.0"),
                ]
            ),
        )

    def test_parse_requirements_with_empty_lines(self):
        self.assertEqual(
            Requirements.from_text("\nfoo\n\nbar==1.0.0\n"),
            Requirements(
                libraries=[
                    Requirement(name="foo"),
                    Requirement(name="bar", version="1.0.0"),
                ]
            ),
        )

    def test_parse_requirements_bad_lines(self):
        self.assertEqual(
            Requirements.from_text("foo\nbar==1.0.0\nas9sdkjcb##8du3789====3ng8ds7\n"),
            Requirements(
                libraries=[
                    Requirement(name="foo"),
                    Requirement(name="bar", version="1.0.0"),
                ]
            ),
        )

    def test_to_dict(self):
        self.assertEqual(
            Requirements(
                libraries=[
                    Requirement(name="foo"),
                    Requirement(name="bar", version="1.0.0"),
                ]
            ).to_dict(),
            [
                {"installed_version": None, "name": "foo", "version": None},
                {"installed_version": None, "name": "bar", "version": "1.0.0"},
            ],
        )

    def test_from_dict(self):
        self.assertEqual(
            Requirements.from_dict(
                [{"name": "foo", "version": None}, {"name": "bar", "version": "1.0.0"}]
            ),
            Requirements(
                libraries=[
                    Requirement(name="foo"),
                    Requirement(name="bar", version="1.0.0"),
                ]
            ),
        )

    def test_add(self):
        requirements = Requirements(libraries=[])
        requirements.add("foo")
        self.assertEqual(requirements.libraries, [Requirement(name="foo")])
        requirements.add("bar", "1.0.0")
        self.assertEqual(
            requirements.libraries,
            [Requirement(name="foo"), Requirement(name="bar", version="1.0.0")],
        )

    def test_add_no_duplicates(self):
        requirements = Requirements(libraries=[Requirement(name="foo")])
        requirements.add("foo")
        self.assertEqual(requirements.libraries, [Requirement(name="foo")])

    def test_valid_has_without_version(self):
        requirements = Requirements(libraries=[Requirement(name="foo")])
        self.assertTrue(requirements.has("foo"))

    def test_valid_has_with_version(self):
        requirements = Requirements(
            libraries=[Requirement(name="foo", version="1.0.0")]
        )
        self.assertTrue(requirements.has("foo", "1.0.0"))

    def test_invalid_has_without_lib(self):
        requirements = Requirements(libraries=[Requirement(name="foo")])
        self.assertFalse(requirements.has("bar"))

    def test_invalid_has_with_wrong_version(self):
        requirements = Requirements(
            libraries=[Requirement(name="foo", version="1.0.0")]
        )
        self.assertFalse(requirements.has("foo", "1.0.1"))

    def test_valid_ensure_with_version_from_specified_version(self):
        requirements = Requirements(
            libraries=[Requirement(name="foo", version="1.0.0")]
        )
        requirements.ensure("foo", "1.0.0")
        self.assertEqual(
            requirements.libraries, [Requirement(name="foo", version="1.0.0")]
        )

    def test_valid_ensure_with_version_from_non_specified_version(self):
        requirements = Requirements(libraries=[Requirement(name="foo")])
        requirements.ensure("foo", "1.0.1")
        self.assertEqual(
            requirements.libraries, [Requirement(name="foo", version="1.0.1")]
        )

    def test_valid_ensure_without_version_from_specified_version(self):
        requirements = Requirements(
            libraries=[Requirement(name="foo", version="1.0.0")]
        )
        requirements.ensure("foo")
        self.assertEqual(
            requirements.libraries, [Requirement(name="foo", version="1.0.0")]
        )

    def test_valid_ensure_without_version_from_non_specified_version(self):
        requirements = Requirements(libraries=[Requirement(name="foo")])
        requirements.ensure("foo")
        self.assertEqual(requirements.libraries, [Requirement(name="foo")])

    def test_valid_update_with_specified_version(self):
        requirements = Requirements(
            libraries=[Requirement(name="foo", version="0.0.0")]
        )
        requirements.update("foo", "0.0.1")
        self.assertEqual(
            requirements.libraries, [Requirement(name="foo", version="0.0.1")]
        )

    def test_valid_update_without_specified_version(self):
        requirements = Requirements(libraries=[Requirement(name="foo")])
        requirements.update("foo", "0.0.1")
        self.assertEqual(
            requirements.libraries, [Requirement(name="foo", version="0.0.1")]
        )

    def test_invalid_update_when_not_found(self):
        requirements = Requirements(libraries=[Requirement(name="foo")])
        requirements.update("bar", "0.0.1")
        self.assertEqual(requirements.libraries, [Requirement(name="foo")])

    def test_is_installed_when_installed(self):
        req = Requirement(name="requests")
        self.assertIsNotNone(req.installed_version())

    def test_is_installed_when_not_installed(self):
        req = Requirement(name="not_installed")
        self.assertIsNone(req.installed_version())

    def test_is_installed_when_installed_with_different_version(self):
        req = Requirement(name="requests", version="0.0.1234")
        self.assertNotEqual(req.installed_version(), req.version)



from typing import Dict, Optional, Union

from abstra_internals.controllers.sdk import WorkflowSDKController
from abstra_internals.repositories import execution_repository, stage_run_repository
from abstra_internals.repositories.project.project import ProjectRepository


def get_data(key: Optional[str] = None) -> Union[Dict[str, object], object]:
    sdk_controller = WorkflowSDKController(
        stage_run_repository=stage_run_repository,
        execution_repository=execution_repository,
        project_repository=ProjectRepository(),
    )
    return sdk_controller.get_data(key)


def set_data(key: str, value: object):
    sdk_controller = WorkflowSDKController(
        stage_run_repository=stage_run_repository,
        execution_repository=execution_repository,
        project_repository=ProjectRepository(),
    )
    sdk_controller.set_data(key, value)


def set_title(title: str):
    set_data("_thread_title", title)



from unittest import TestCase

from abstra_internals.constants import get_persistent_dir
from abstra_internals.utils.dot_abstra import PERSISTENT_DIR
from tests.fixtures import clear_dir, init_dir


class GetPersistentDirTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_value(self):
        self.assertFalse((self.root / PERSISTENT_DIR).exists())
        self.assertEqual(self.root / PERSISTENT_DIR, get_persistent_dir())
        self.assertTrue((self.root / PERSISTENT_DIR).exists())



import base64
import binascii
import io
import pathlib
import re
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from PIL.Image import Image
import puremagic

from abstra_internals.utils.file import get_random_filepath

BASE_64_PREFIX = "data:image/png;base64,"
PREFFIX_REGEX = re.compile(r"^data:image/(jpeg|png|gif|bmp|webp);base64,")


def is_base_64(unknown_string: str) -> bool:
    if PREFFIX_REGEX.match(unknown_string):
        unknown_string = unknown_string.split(",")[1]
    try:
        base64_bytes = base64.b64decode(unknown_string, validate=True)
        image_type = puremagic.what(None, h=base64_bytes)
        if image_type in ["jpeg", "png", "gif", "bmp", "webp"]:
            return True
        else:
            return False
    except (binascii.Error, ValueError):
        return False


def to_base64(file: Union[str, io.IOBase, pathlib.Path, "Image"]) -> Union[str, None]:
    if not file:
        return ""

    if isinstance(file, pathlib.Path):
        with file.open("rb") as f:
            file_content = f.read()
            base64_file = base64.b64encode(file_content).decode("utf-8")
            return f"{BASE_64_PREFIX}{base64_file}"

    if isinstance(file, io.IOBase):
        file.seek(0)
        file_content = file.read()
        base64_file = base64.b64encode(file_content).decode("utf-8")
        return f"{BASE_64_PREFIX}{base64_file}"

    if isinstance(file, str):
        # URL or base64 encoded string
        if file.startswith("http") or file.startswith("data:"):
            return file

        # path to file
        return f"{BASE_64_PREFIX}{file}"

    from PIL.Image import Image

    if isinstance(file, Image):
        _, file_path = get_random_filepath()
        file.save(str(file_path))
        with file_path.open("rb") as f:
            file_content = f.read()
            base64_file = base64.b64encode(file_content).decode("utf-8")
            return f"{BASE_64_PREFIX}{base64_file}"

    # FileResponse. TODO: check with isinstance without circular import
    if hasattr(file, "file"):
        if isinstance(file.file, io.IOBase):
            file.file.seek(0)
            file_content = file.file.read()
            base64_file = base64.b64encode(file_content).decode("utf-8")
            return f"{BASE_64_PREFIX}{base64_file}"

    return None



import flask

from abstra_internals.controllers.main import MainController
from abstra_internals.server.routes import access_control as ac_router
from abstra_internals.server.routes import ai as ai_router
from abstra_internals.server.routes import assets as assets_router
from abstra_internals.server.routes import debugger as debugger_router
from abstra_internals.server.routes import env_vars as envvars_router
from abstra_internals.server.routes import forms as forms_router
from abstra_internals.server.routes import hooks as hooks_router
from abstra_internals.server.routes import jobs as jobs_router
from abstra_internals.server.routes import kanban as kanban_router
from abstra_internals.server.routes import linters as linters_router
from abstra_internals.server.routes import login as login_router
from abstra_internals.server.routes import pysa as pysa_router
from abstra_internals.server.routes import requirements as requirements_router
from abstra_internals.server.routes import roles as roles_router
from abstra_internals.server.routes import scripts as scripts_router
from abstra_internals.server.routes import stage_runs as stage_runs_router
from abstra_internals.server.routes import stdio as stdio_router
from abstra_internals.server.routes import workflows as workflows_router
from abstra_internals.server.routes import workspace as workspace_router
from abstra_internals.server.utils import send_from_dist
from abstra_internals.usage import editor_usage


def __get_api_bp(controller: MainController):
    bp = flask.Blueprint("editor_api", __name__)

    kanban_bp = kanban_router.get_editor_bp(controller)
    bp.register_blueprint(kanban_bp, url_prefix="/kanban")

    stage_run_bp = stage_runs_router.get_editor_bp(controller)
    bp.register_blueprint(stage_run_bp, url_prefix="/stage_runs")

    workspace_bp = workspace_router.get_editor_bp(controller)
    bp.register_blueprint(workspace_bp, url_prefix="/workspace")

    forms_bp = forms_router.get_editor_bp(controller)
    bp.register_blueprint(forms_bp, url_prefix="/forms")

    hooks_bp = hooks_router.get_editor_bp(controller)
    bp.register_blueprint(hooks_bp, url_prefix="/hooks")

    jobs_bp = jobs_router.get_editor_bp(controller)
    bp.register_blueprint(jobs_bp, url_prefix="/jobs")

    scripts_bp = scripts_router.get_editor_bp(controller)
    bp.register_blueprint(scripts_bp, url_prefix="/scripts")

    workflows_bp = workflows_router.get_editor_bp(controller)
    bp.register_blueprint(workflows_bp, url_prefix="/workflows")

    requirements_bp = requirements_router.get_editor_bp(controller)
    bp.register_blueprint(requirements_bp, url_prefix="/requirements")

    debugger_bp = debugger_router.get_editor_bp(controller)
    bp.register_blueprint(debugger_bp, url_prefix="/debugger")

    login_bp = login_router.get_editor_bp(controller)
    bp.register_blueprint(login_bp, url_prefix="/login")

    linters_bp = linters_router.get_editor_bp()
    bp.register_blueprint(linters_bp, url_prefix="/linters")

    ai_bp = ai_router.get_editor_bp(controller)
    bp.register_blueprint(ai_bp, url_prefix="/ai")

    assets_bp = assets_router.get_editor_bp(controller)
    bp.register_blueprint(assets_bp, url_prefix="/assets")

    envvars_bp = envvars_router.get_editor_bp(controller)
    bp.register_blueprint(envvars_bp, url_prefix="/env-vars")

    pysa_bp = pysa_router.get_editor_bp()
    bp.register_blueprint(pysa_bp, url_prefix="/pysa")

    roles_bp = roles_router.get_editor_bp(controller)
    bp.register_blueprint(roles_bp, url_prefix="/roles")

    access_control_bp = ac_router.get_editor_bp(controller)
    bp.register_blueprint(access_control_bp, url_prefix="/access-control")

    stdio_bp = stdio_router.get_editor_bp(controller)
    bp.register_blueprint(stdio_bp, url_prefix="/stdio")

    return bp


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor", __name__)

    api_bp = __get_api_bp(controller)
    bp.register_blueprint(api_bp, url_prefix="/api")

    @bp.get("/")
    @editor_usage
    def _spa_index():
        return send_from_dist("editor.html", "editor.html")

    @bp.get("/<path:filename>")
    def _spa(filename: str):
        return send_from_dist(filename, "editor.html")

    return bp



import json

from abstra_internals.settings import Settings
from abstra_internals.templates import vscode_launch_json
from abstra_internals.utils import serialize

CONNECTED_PORT = None


def is_launch_json_configured():
    try:
        launch_json_file = Settings.root_path.joinpath(".vscode/launch.json")
        if not launch_json_file.exists():
            return False

        launch_json = json.loads(launch_json_file.read_text(encoding="utf-8"))
        if launch_json["version"] != "0.2.0":
            return False

        configurations = launch_json["configurations"]
        if len(configurations) == 0:
            return False
        configuration = [c for c in configurations if c["name"] == "Abstra debugger"][0]
        if configuration["type"] != "python":
            return False
        if configuration["request"] != "attach":
            return False
        if configuration["connect"]["host"] not in [
            "localhost",
            "127.0.0.1",
            "0.0.0.0",
        ]:
            return False
        if configuration["connect"]["port"] != "${input:abstraDebugPort}":
            return False
        inputs = launch_json["inputs"]
        if len(inputs) == 0:
            return False
        input = [i for i in inputs if i["id"] == "abstraDebugPort"][0]
        if input["type"] != "promptString":
            return False
        if input["id"] != "abstraDebugPort":
            return False
        return True
    except Exception:
        return False


def configure_launch_json():
    if is_launch_json_configured():
        return
    Settings.root_path.joinpath(".vscode").mkdir(exist_ok=True)
    Settings.root_path.joinpath(".vscode/launch.json").write_text(
        serialize(vscode_launch_json, indent=4), encoding="utf-8"
    )



import tempfile
import unittest
from pathlib import Path

from abstra_internals.interface.cli.start import start
from abstra_internals.settings import SettingsController
from tests.fixtures import clear_dir, init_dir


class TestStart(unittest.TestCase):
    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_project_create_files_correctly(self):
        self.root = Path(tempfile.mkdtemp())
        SettingsController.set_root_path(self.root.as_posix())

        start()
        self.assertTrue((self.root / "abstra.json").exists())
        self.assertTrue((self.root / ".abstraignore").exists())
        self.assertTrue((self.root / ".gitignore").exists())
        self.assertTrue((self.root / "application_form.py").exists())
        self.assertTrue((self.root / "application_review.py").exists())

    def test_dont_generate_project_if_there_is_a_project_in_directory(self):
        self.root = init_dir()

        self.assertTrue((self.root / "abstra.json").exists())

        start()
        self.assertFalse((self.root / "application_form.py").exists())

    def test_create_project_with_credentials(self):
        self.root = Path(tempfile.mkdtemp())
        SettingsController.set_root_path(self.root.as_posix())

        start(credential_token="test_token")
        self.assertTrue((self.root / ".abstra" / "credentials").exists())
        self.assertEqual(
            (self.root / ".abstra" / "credentials").read_text(), "test_token"
        )



import flask

from abstra_internals.controllers.main import MainController
from abstra_internals.jwt_auth import USER_AUTH_HEADER_KEY, UserClaims
from abstra_internals.usage import player_usage


def get_player_bp(main_controller: MainController):
    bp = flask.Blueprint("playerUsers", __name__)

    @bp.post("/sign-up")
    @player_usage
    def _signup_user():
        auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
        if auth is None:
            flask.abort(401)

        claims = UserClaims.from_jwt(auth.split(" ")[1])
        if claims is None or claims.email is None:
            flask.abort(401)

        success = main_controller.users_repository.insert_user(claims.email)

        return {"success": success}

    return bp



import ast
import fnmatch
import os
import re
import tempfile
import uuid
import warnings
from pathlib import Path
from typing import Generator, Optional, Union

from abstra_internals.utils.dot_abstra import DOT_ABSTRA_FOLDER_NAME

GIT_FOLDER = ".git"
ABSTRA_IGNORE = ".abstraignore"
ABSTRA_TABLES_FILE = "abstra-tables.json"


def is_relative_path(path: str) -> bool:
    return not path.startswith("<") and not path.startswith("/")


def get_ignore_files(dir: Path):
    GITIGNORE = dir.joinpath(".gitignore")
    IGNOREFILE = dir.joinpath(ABSTRA_IGNORE)
    git_path = dir.joinpath(GIT_FOLDER)
    abstra_path = dir.joinpath(DOT_ABSTRA_FOLDER_NAME)
    ignored = [IGNOREFILE.name, GITIGNORE.name, abstra_path.name, git_path.name]

    if IGNOREFILE.exists():
        with open(IGNOREFILE, "r", encoding="utf-8"):
            ignored.extend(IGNOREFILE.read_text(encoding="utf-8").split("\n"))
    elif GITIGNORE.exists():
        ignored.extend(GITIGNORE.read_text(encoding="utf-8").split("\n"))
    return [
        re.compile(fnmatch.translate(f"{p}*"))
        for p in ignored
        if p and not p.startswith("#") and not p.startswith("!")
    ]


def should_ignore(ignored_paths, _path: Path):
    path = str(_path)
    for ignored_path in ignored_paths:
        if ignored_path.search(path):
            return True
    return False


def files_from_directory(directory: Path):
    ignored = [*get_ignore_files(directory), *get_ignore_files(Path.cwd())]
    paths = filter(
        lambda p: p.is_file() and not should_ignore(ignored, p.relative_to(directory)),
        Path(directory).rglob("*"),
    )
    return list(paths)


def path2module(path: Path) -> str:
    if path.suffix == ".py":
        path = path.with_suffix("")
    if path.name == "__init__":
        path = path.parent
    return str(path).replace("/", ".").replace("\\", ".")


def module2path(module: str, package: bool) -> Path:
    path_list = module.split(".")

    # Handle trailing dots
    is_first_trailing_dot = True
    for index, path in enumerate(path_list):
        if path.__len__() == 0:
            if is_first_trailing_dot:
                path_list[index] = "."
                is_first_trailing_dot = False
            else:
                path_list[index] = ".."
        else:
            break

    if package:
        path_list.append("__init__.py")
    else:
        path_list[-1] += ".py"
    return Path(*path_list)


def _get_file_path_from_absolute_module(module: str) -> Union[Path, None]:
    module_path = Path(module2path(module, package=False))
    if module_path.is_file():
        return module_path

    package_path = Path(module2path(module, package=True))
    if package_path.is_file():
        return package_path

    return None


def _get_file_path_from_relative_module(
    module: str, parent_path: Path
) -> Union[Path, None]:
    module_path = parent_path / module2path(module, package=False)
    resolved_module_path = module_path.resolve().relative_to(os.getcwd())

    if resolved_module_path.is_file():
        return resolved_module_path

    package_path = parent_path / module2path(module, package=True)
    resolved_package_path = package_path.resolve().relative_to(os.getcwd())

    if resolved_package_path.is_file():
        return resolved_package_path

    return None


def _get_file_path_depending_on_node_level(
    node_level: int, module: str, path_parent: Path
) -> Optional[Path]:
    return (
        _get_file_path_from_relative_module(module, path_parent)
        if node_level > 0
        else _get_file_path_from_absolute_module(module)
    )


def _get_file_path(
    node_level: int,
    module: Optional[str],
    alias_name: str,
    path_parent: Path,
) -> Optional[Path]:
    prefix = "." * (node_level)
    has_module = module is not None and module != ""

    # Check if it's importing a file
    module_name = f"{module}." if has_module else ""
    relative_module = f"{prefix}{module_name}{alias_name}"
    file_path = _get_file_path_depending_on_node_level(
        node_level, relative_module, path_parent
    )

    # Check if it's importing an entity from a file
    if file_path is None:
        relative_module = f"{prefix}{module if has_module else ''}"
        file_path = _get_file_path_depending_on_node_level(
            node_level, relative_module, path_parent
        )

    return file_path


def silent_traverse_code(
    path: Path, raise_on_syntax_errors=False
) -> Generator[Path, None, None]:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        yield from traverse_code(path, raise_on_syntax_errors)


def traverse_code(
    path: Path, raise_on_syntax_errors=False
) -> Generator[Path, None, None]:
    yield path

    try:
        code = path.read_text(encoding="utf-8")
        parsed = ast.parse(code)

        for node in parsed.body:
            if isinstance(node, ast.Import):
                for alias in node.names:
                    file_path = _get_file_path_from_absolute_module(alias.name)

                    if file_path is not None:
                        yield from traverse_code(file_path)

            if isinstance(node, ast.ImportFrom):
                for alias in node.names:
                    file_path = _get_file_path(
                        node.level, node.module, alias.name, path.parent
                    )

                    if file_path is not None:
                        yield from traverse_code(file_path)

    except Exception as e:
        if raise_on_syntax_errors and isinstance(e, SyntaxError):
            raise e


def get_tmp_upload_dir():
    uploaded_files_dir = Path(tempfile.gettempdir(), "_uploaded_files")
    uploaded_files_dir.mkdir(exist_ok=True)
    return uploaded_files_dir


def get_random_filepath(name=None):
    if name is None:
        name = str(uuid.uuid4())
    else:
        name = str(uuid.uuid4()) + "/" + Path(name).name
    path = get_tmp_upload_dir() / name
    path.parent.mkdir(exist_ok=True)
    return name, path


CONFLICTING_STATIC_PATHS = [
    # Player routes
    "oidc-login-callback",
    "oidc-logout-callback",
    "login",
    "threads",
]

CONFLICTING_DYNAMIC_PATHS = [
    # Player routes
    "error/",
    "_player/",
    "form/",
    "forms/",
    "hook/",
    "hooks/",
    # Editor routes
    "_editor/",
]


def is_path_in_conflict(path: str):
    return path in CONFLICTING_STATIC_PATHS or any(
        path.startswith(p) for p in CONFLICTING_DYNAMIC_PATHS
    )


def generate_conflictless_path(path: str):
    if is_path_in_conflict(path):
        random_suffix = str(uuid.uuid4())[:8]
        return f"{path}-{random_suffix}"
    else:
        return path



from abstra_internals.interface.sdk.connectors.google_credentials import (
    create_abstra_google_credentials,
)
from abstra_internals.repositories import connectors_repository
from abstra_internals.repositories.connectors import AccessTokenDTO


def get_access_token(name: str) -> AccessTokenDTO:
    return connectors_repository.get_access_token(name)


def get_gsheets_credentials(connection_name: str = "google-sheets"):
    return create_abstra_google_credentials(connectors_repository, connection_name)


def get_gdrive_credentials(connection_name: str = "google-drive"):
    return create_abstra_google_credentials(connectors_repository, connection_name)


def get_gmail_credentials(connection_name: str = "gmail"):
    return create_abstra_google_credentials(connectors_repository, connection_name)


def get_bigquery_credentials(connection_name: str = "google-bigquery"):
    return create_abstra_google_credentials(connectors_repository, connection_name)


def get_google_cal_credentials(connection_name: str = "google-calendar"):
    return create_abstra_google_credentials(connectors_repository, connection_name)



from pathlib import Path
from unittest import TestCase

from abstra_internals.controllers.main import MainController
from abstra_internals.linter.rules.syntax_errors import SyntaxErrors
from tests.fixtures import clear_dir, init_dir


class SyntaxErrorsTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_syntax_errors_valid_default(self):
        rule = SyntaxErrors()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_syntax_errors_valid_empty_file(self):
        script = MainController().create_script("New script", "script.py")
        script.file_path.touch()
        rule = SyntaxErrors()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_syntax_errors_valid_with_content(self):
        script = MainController().create_script("New script", "script.py")
        script.file_path.write_text("print('hello world')")
        rule = SyntaxErrors()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_syntax_errors_invalid_with_syntax_error(self):
        script = MainController().create_script("New script", "script.py")
        script.file_path.write_text("print('hello world'")
        rule = SyntaxErrors()
        self.assertEqual(len(rule.find_issues()), 1)
        assert rule.error is not None
        assert rule.error.filename is not None
        self.assertEqual(Path(script.file).name, Path(rule.error.filename).name)



import flask

from abstra_internals.controllers.main import MainController
from abstra_internals.controllers.stage_runs import StageRunsController
from abstra_internals.environment import IS_PRODUCTION
from abstra_internals.repositories.stage_run import (
    GetStageRunByQueryFilter,
    Pagination,
)
from abstra_internals.server.guards.role_guard import Guard, StageIdSelector
from abstra_internals.usage import editor_usage, player_usage


def get_editor_bp(main_controller: MainController):
    bp = flask.Blueprint("editor_stage_runs", __name__)
    controller = StageRunsController(main_controller.stage_run_repository)

    @bp.get("/")
    @editor_usage
    def _get_leaf_stage_runs():
        filter = GetStageRunByQueryFilter.from_dict(flask.request.args)
        pagination = Pagination.from_dict(flask.request.args)
        return controller.get_leaf_stage_runs(filter, pagination)

    @bp.get("/past")
    @editor_usage
    def _get_past_stage_runs():
        filter = GetStageRunByQueryFilter.from_dict(flask.request.args)
        pagination = Pagination.from_dict(flask.request.args)
        return controller.get_past_stage_runs(filter, pagination)

    @bp.post("/fork")
    @editor_usage
    def _fork():
        data = flask.request.json
        if not data:
            flask.abort(400)
        stage_run_id = data.get("stage_run_id")
        custom_thread_data = data.get("custom_thread_data")
        if not stage_run_id:
            flask.abort(400)
        return controller.fork(stage_run_id, custom_thread_data)

    @bp.post("/retry")
    @editor_usage
    def _retry():
        data = flask.request.json
        if not data:
            flask.abort(400)
        stage_run_id = data.get("stage_run_id")
        if not stage_run_id:
            flask.abort(400)
        return controller.retry(stage_run_id)

    return bp


def get_player_bp(main_controller: MainController):
    bp = flask.Blueprint("player_stage_runs", __name__)
    controller = StageRunsController(main_controller.stage_run_repository)

    guard = Guard(main_controller.users_repository, enabled=IS_PRODUCTION)

    @bp.post("/retry")
    @guard.by(StageIdSelector("kanban"))
    @player_usage
    def _retry():
        data = flask.request.json
        if not data:
            flask.abort(400)
        stage_run_id = data.get("stage_run_id")
        if not stage_run_id:
            flask.abort(400)
        return controller.retry(stage_run_id)

    return bp



from typing import Dict, Optional

from abstra_internals.controllers.sdk import FormSDKController
from abstra_internals.jwt_auth import UserClaims
from abstra_internals.proxy import ReadOnlyProxyDict
from abstra_internals.repositories import users_repository


def get_user(refresh: bool = False) -> UserClaims:
    """
    Get the current user

    Args:
        refresh (bool, optional): Whether to force refresh the user. Defaults to False.

    Returns:
        Union[UserClaims, None]: User information
    """

    sdk_controller = FormSDKController(users_repository=users_repository)
    return sdk_controller.get_user(force_refresh=refresh)


def execute_js(code: str, context: Optional[dict] = None):
    """
    Execute JavaScript code in the browser

    Args:
        code (str): JavaScript code to execute
        context (dict, optional): Context to pass to the JavaScript code. Defaults to {}.

    Returns:
        Any: Result of the JavaScript code
    """

    sdk_controller = FormSDKController(users_repository=users_repository)
    return sdk_controller.execute_js(code, context or {})


def redirect(url: str, query_params: Optional[dict] = None) -> None:
    """
    Redirect the user to a new URL

    Args:
        url (str): URL to redirect to
        query_params (Optional[dict], optional): Query parameters to append to the URL. Defaults to None.

    Returns:
        None
    """
    sdk_controller = FormSDKController(users_repository=users_repository)
    sdk_controller.redirect(url, query_params)


def get_query_params() -> Dict[str, str]:
    """
    Get the query parameters from the current URL

    Returns:
        Dict[str, str]: Query parameters
    """

    sdk_controller = FormSDKController(users_repository=users_repository)
    return sdk_controller.get_query_params()


# legacy
url_params = ReadOnlyProxyDict(get_query_params)


__all__ = [
    "get_user",
    "execute_js",
    "redirect",
    "get_query_params",
    "url_params",
]



import logging
import threading

from dotenv import load_dotenv as _load_dotenv
from werkzeug.debug import DebuggedApplication
from werkzeug.serving import make_server

from abstra_internals.controllers.main import MainController
from abstra_internals.editor_reloader import LocalReloader
from abstra_internals.environment import HOST
from abstra_internals.fs_watcher import files_changed_polling_loop
from abstra_internals.interface.cli.messages import serve_message
from abstra_internals.server.apps import get_local_app
from abstra_internals.settings import Settings
from abstra_internals.stdio_patcher import StdioPatcher
from abstra_internals.utils.browser import browser_open_editor
from abstra_internals.version import check_latest_version


def editor(
    debug: bool,
    load_dotenv: bool,
    reloading: bool,
):
    serve_message()
    check_latest_version()

    controller = MainController()
    controller.reset_repositories()
    StdioPatcher.apply(controller)

    # watch_py_root_files
    threading.Thread(
        daemon=True,
        name="file_watcher",
        target=files_changed_polling_loop,
        kwargs=dict(controller=controller),
    ).start()

    if load_dotenv:
        _load_dotenv(Settings.root_path / ".env")

    app = get_local_app(controller)
    if not debug:
        log = logging.getLogger("werkzeug")
        log.setLevel(logging.WARNING)
    else:
        app = DebuggedApplication(app, evalex=True)

    server = make_server(host=HOST, port=Settings.server_port, threaded=True, app=app)

    LocalReloader.set_server(server)

    if not reloading:
        browser_open_editor()

    server.serve_forever()



import json
import threading
from pathlib import Path
from typing import Generic, List, Optional, Protocol, Type, TypeVar

from abstra_internals.settings import Settings


class Serializable(Protocol):
    def to_dto(self) -> dict: ...

    @staticmethod
    def from_dto(dto: dict) -> "Serializable": ...


T = TypeVar("T", bound=Serializable)


class FileManager(Generic[T]):
    locks = {}

    @classmethod
    def _get_or_create_lock(cls, directory: str) -> threading.RLock:
        lock = cls.locks.get(directory)
        if lock is None:
            lock = threading.RLock()
            cls.locks[directory] = lock
        return lock

    def __init__(self, directory: str, model: Type[T]):
        self.directory = directory
        self.model = model
        self.lock = FileManager._get_or_create_lock(self.directory)

    @property
    def directory_path(self) -> Path:
        return Settings.root_path / self.directory

    def save(self, id: str, data: T) -> None:
        with self.lock:
            file_path = self._get_file_path(id)

            with file_path.open("w", encoding="utf-8") as f:
                json.dump(data.to_dto(), f, indent=4)

    def load_all(self) -> List[T]:
        with self.lock:
            data = [
                self._load(file_path.stem)
                for file_path in self.directory_path.glob("*.json")
            ]

            return [d for d in data if d is not None]

    def load(self, id: str) -> Optional[T]:
        with self.lock:
            return self._load(id)

    def delete(self, id: str) -> None:
        with self.lock:
            file_path = self._get_file_path(id)
            if file_path.exists():
                file_path.unlink()

    def clear(self) -> None:
        with self.lock:
            for file_path in self.directory_path.glob("*.json"):
                file_path.unlink()

    def _get_file_path(self, id: str) -> Path:
        self.directory_path.mkdir(parents=True, exist_ok=True)
        return self.directory_path / f"{id}.json"

    def _load(self, id: str) -> Optional[T]:
        file_path = self._get_file_path(id)
        if file_path.exists():
            with file_path.open("r", encoding="utf-8") as f:
                try:
                    dto = json.load(f)
                    return self.model.from_dto(dto)
                except Exception:
                    self.delete(id)

        return None



import os
import time
from datetime import datetime

from abstra_internals.controllers.main import MainController
from abstra_internals.modules import reload_project_local_modules


def has_local_dependencies_changed(
    controller: MainController, last_change: float
) -> bool:
    for file in controller.get_project_local_dependencies():
        if not file.exists():
            continue

        st_mtime = os.stat(file).st_mtime
        if st_mtime > last_change:
            return True
    return False


def reload_files_on_change(controller: MainController, last_change: float):
    if not has_local_dependencies_changed(controller, last_change):
        return False

    try:
        reload_project_local_modules()
        return True
    except Exception:
        return False


def files_changed_polling_loop(*, controller: MainController):
    last_change = datetime.now().timestamp()

    while True:
        try:
            has_reloaded = reload_files_on_change(controller, last_change)
            if has_reloaded:
                last_change = datetime.now().timestamp()
        except Exception:
            pass
        time.sleep(1)



import json
from typing import Any
from abstra_internals.widgets.widget_base import Output


class PlotlyOutput(Output):
    type = 'plotly-output'

    def __init__(self, fig: Any=None, **kwargs):
        self.set_props(dict(fig=fig, **kwargs))

    def set_props(self, props):
        self.fig = props.get('fig', None)
        self.full_width = props.get('full_width', False)
        self.label = props.get('label', '')

    def serialize_figure(self):
        if self.fig is None:
            import plotly.express as px
            df = px.data.tips()
            fig = px.density_heatmap(df, x='total_bill', y='tip')
            fig_json = fig.to_json()
            if not isinstance(fig_json, str):
                raise Exception('fig.to_json() did not return a string')
            return json.loads(fig_json)
        return json.loads(self.fig.to_json())

    def render(self, ctx: dict):
        return {'type': self.type, 'figure': self.serialize_figure(),
            'fullWidth': self.full_width, 'label': self.label}



from pathlib import Path
from unittest import TestCase

from abstra_internals.controllers.execution import ExecutionController
from abstra_internals.controllers.execution_client_hook import HookClient
from abstra_internals.controllers.workflow_engine import WorkflowEngine
from abstra_internals.controllers.workflow_engine_detached import DetachedWorkflowEngine
from abstra_internals.entities.execution import RequestContext
from abstra_internals.repositories.email import TestEmailRepository
from abstra_internals.repositories.execution import EditorExecutionRepository
from abstra_internals.repositories.execution_logs import LocalExecutionLogsRepository
from abstra_internals.repositories.project.project import (
    HookStage,
    ProjectRepository,
)
from abstra_internals.repositories.stage_run import LocalStageRunRepository
from tests.fixtures import clear_dir, init_dir


class ExecutionControllerTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()
        self.stage_run_repository = LocalStageRunRepository()
        self.execution_repository = EditorExecutionRepository()

        self.workflow_engine = WorkflowEngine(
            stage_run_repository=self.stage_run_repository,
            execution_repository=self.execution_repository,
            email_repository=TestEmailRepository(),
            execution_logs_repository=LocalExecutionLogsRepository(),
        )

        self.detached_workflow_engine = DetachedWorkflowEngine(
            stage_run_repository=self.stage_run_repository
        )

        self.request_context = RequestContext(
            body={"a": 1}.__str__(),
            headers={"auth": "some_secret_token"},
            query_params={"c": "3"},
            method="GET",
        )

        self.project = ProjectRepository.load()
        self.stage = HookStage.create(
            title="mock_stage",
            file="mock_file.py",
            id="mock_hook_id",
            workflow_position=(0, 1),
        )
        Path(self.stage.file).write_text("print('Hello, World!')")
        self.project.add_stage(self.stage)
        ProjectRepository.save(self.project)

        self.hook_client = HookClient(self.request_context)

        return super().setUp()

    def tearDown(self) -> None:
        clear_dir(self.root)
        return super().tearDown()

    def test_run_initial_returns_dto(self):
        ExecutionController(
            workflow_engine=self.workflow_engine,
            stage_run_repository=self.stage_run_repository,
            execution_repository=self.execution_repository,
        ).run(
            stage=self.stage,
            client=self.hook_client,
            request=self.request_context,
        )

        self.assertEqual(self.hook_client.response["status"], 200)

    def test_run_detached_returns_dto(self):
        ExecutionController(
            workflow_engine=self.detached_workflow_engine,
            stage_run_repository=self.stage_run_repository,
            execution_repository=self.execution_repository,
        ).test(
            stage=self.stage,
            client=self.hook_client,
            request=self.request_context,
        )

        self.assertEqual(self.hook_client.response["status"], 200)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Literal, Optional, Tuple, Union, get_args

CONDITION_COMPARATORS_TYPE = Literal["is", "contains"]

SINGLE_CONDITION_OPERATORS_TYPE = Literal["NOT"]
SINGLE_CONDITION_OPERATORS: Tuple[SINGLE_CONDITION_OPERATORS_TYPE, ...] = get_args(
    SINGLE_CONDITION_OPERATORS_TYPE
)

MULTIPLE_CONDITIONS_OPERATORS_TYPE = Literal["AND", "OR"]
MULTIPLE_CONDITIONS_OPERATORS: Tuple[MULTIPLE_CONDITIONS_OPERATORS_TYPE, ...] = (
    get_args(MULTIPLE_CONDITIONS_OPERATORS_TYPE)
)


@dataclass
class FilterCondition:
    @staticmethod
    def from_dict(data: dict) -> Optional[FilterCondition]:
        if "key" in data and "comparator" in data and "value" in data:
            if data["key"] == "":
                return None
            return Condition(
                key=data["key"], comparator=data["comparator"], value=data["value"]
            )
        if "operator" in data:
            if data["operator"] in SINGLE_CONDITION_OPERATORS:
                return LogicalGroupSingleCondition(
                    operator=data["operator"],
                    condition=FilterCondition.from_dict(data["condition"]),
                )
            if data["operator"] in MULTIPLE_CONDITIONS_OPERATORS:
                return LogicalGroupMultipleConditions(
                    operator=data["operator"],
                    conditions=[
                        FilterCondition.from_dict(condition)
                        for condition in data["conditions"]
                    ],
                )

        return None

    def to_dict(self) -> dict:
        if isinstance(self, Condition):
            return {
                "key": self.key,
                "comparator": self.comparator,
                "value": self.value,
            }
        if isinstance(self, LogicalGroupSingleCondition):
            return {
                "operator": self.operator,
                "condition": self.condition.to_dict() if self.condition else None,
            }
        if isinstance(self, LogicalGroupMultipleConditions):
            return {
                "operator": self.operator,
                "conditions": [
                    condition.to_dict() if condition else None
                    for condition in (self.conditions or [])
                ],
            }

        return {}


@dataclass
class Condition(FilterCondition):
    key: str
    comparator: CONDITION_COMPARATORS_TYPE
    value: str


@dataclass
class LogicalGroupSingleCondition(FilterCondition):
    operator: SINGLE_CONDITION_OPERATORS_TYPE
    condition: Optional[FilterCondition]


@dataclass
class LogicalGroupMultipleConditions(FilterCondition):
    operator: MULTIPLE_CONDITIONS_OPERATORS_TYPE
    conditions: Optional[List[Optional[FilterCondition]]]


LogicalGroup = Union[LogicalGroupSingleCondition, LogicalGroupMultipleConditions]


def evaluate_condition(condition: Condition, data: dict):
    if condition.key == "":
        return True

    if condition.key not in data:
        return False

    if condition.comparator == "is":
        return data[condition.key] == condition.value

    if condition.comparator == "contains":
        if not isinstance(data[condition.key], str):
            return False
        return condition.value in data[condition.key]

    return False


def evaluate_logical_group(group: LogicalGroup, data: dict):
    if group.operator == "AND":
        if group.conditions is None or len(group.conditions) == 0:
            return True
        return all(
            evaluate_filter_condition(condition, data) for condition in group.conditions
        )
    if group.operator == "OR":
        if group.conditions is None or len(group.conditions) == 0:
            return True
        return any(
            evaluate_filter_condition(condition, data) for condition in group.conditions
        )
    if group.operator == "NOT":
        if group.condition is None:
            return True
        return not evaluate_filter_condition(group.condition, data)

    return False


def evaluate_filter_condition(filter_condition: Optional[FilterCondition], data: dict):
    if isinstance(filter_condition, Condition):
        return evaluate_condition(filter_condition, data)
    if isinstance(
        filter_condition, (LogicalGroupSingleCondition, LogicalGroupMultipleConditions)
    ):
        return evaluate_logical_group(filter_condition, data)
    if filter_condition is None:
        return True

    return False



import re
from abstra_internals.widgets.response_types import PhoneResponse
from abstra_internals.widgets.widget_base import Input


class PhoneInput(Input):
    type = 'phone-input'
    empty_value: PhoneResponse = PhoneResponse(raw='', masked='',
        country_code='', national_number='')

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.value = props.get('initial_value', None)
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.placeholder = props.get('placeholder', '')
        self.full_width = props.get('full_width', False)
        self.disabled = props.get('disabled', False)
        self.invalid_message = props.get('invalid_message',
            'i18n_error_invalid_phone_number')

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'value': self.serialize_value(), 'placeholder': self.
            placeholder, 'required': self.required, 'hint': self.hint,
            'fullWidth': self.full_width, 'disabled': self.disabled,
            'errors': self.errors, 'invalidMessage': self.invalid_message}

    def validate(self):
        errors = []
        if self.required and (self.value.country_code == '' or self.value.
            national_number == ''):
            errors.append('i18n_error_required_field')
        return errors

    def is_value_unset(self):
        return self.value.raw == ''

    def set_value(self, value, set_errors=False):
        if isinstance(value, PhoneResponse):
            self.value = value
        elif isinstance(value, dict):
            self.value = self.parse_value(value)
        elif value is None:
            self.value = self.empty_value
        if set_errors:
            self.set_errors()

    def serialize_value(self):
        if isinstance(self.value, dict):
            return {'countryCode': re.sub('\\D', '', str(self.value.get(
                'country_code', ''))), 'nationalNumber': re.sub('\\D', '',
                str(self.value.get('national_number', '')))}
        if isinstance(self.value, PhoneResponse):
            return {'countryCode': re.sub('\\D', '', str(self.value.
                country_code)), 'nationalNumber': re.sub('\\D', '', str(
                self.value.national_number))}
        return {'countryCode': '', 'nationalNumber': ''}

    def parse_value(self, value) ->PhoneResponse:
        if value is None:
            return self.empty_value
        masked = f"+{value.get('countryCode')} {value.get('nationalNumber')}"
        raw = re.sub('\\D', '', masked)
        return PhoneResponse(raw=raw, masked=masked, country_code=
            f"+{value.get('countryCode')}", national_number=value.get(
            'nationalNumber'))



import unittest
from pathlib import Path

from abstra_internals.utils.dot_abstra import DOT_ABSTRA_FOLDER_NAME
from abstra_internals.utils.file import (
    files_from_directory,
    generate_conflictless_path,
    module2path,
    path2module,
)
from tests.fixtures import clear_dir, init_dir


class Module2PathTest(unittest.TestCase):
    def test_module2path_module(self):
        module = "a.b.c"
        path = module2path(module, False)
        self.assertEqual(path, Path("a", "b", "c.py"))

    def test_module2path_package(self):
        module = "a.b.c"
        path = module2path(module, True)
        self.assertEqual(path, Path("a", "b", "c", "__init__.py"))


class Path2ModuleTest(unittest.TestCase):
    def test_path2module_module(self):
        path = Path("a", "b", "c.py")
        module = path2module(path)
        self.assertEqual(module, "a.b.c")

    def test_path2module_package(self):
        path = Path("a", "b", "c", "__init__.py")
        module = path2module(path)
        self.assertEqual(module, "a.b.c")


class FilesFromDirectoryTest(unittest.TestCase):
    def add_file(self, name, content, path=None):
        filepath = (path or self.path).joinpath(name)
        filepath.write_text(content, encoding="utf-8")
        return filepath

    def add_folder(self, name, path=None):
        folderpath = (path or self.path).joinpath(name)
        folderpath.mkdir()
        return folderpath

    def setUp(self) -> None:
        self.path = init_dir()
        self.initial_files = [
            self.path.joinpath("abstra.json"),
            self.path.joinpath("logo.png"),
            self.path.joinpath("favicon.ico"),
        ]

    def tearDown(self) -> None:
        clear_dir(self.path)

    def assertSamePaths(self, paths, expected):
        self.assertEqual(set([str(p) for p in paths]), set([str(p) for p in expected]))

    def test_empty_directory(self):
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [*self.initial_files])

    def test_no_ignore(self):
        filepath = self.add_file("foo", "bar")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [filepath, *self.initial_files])

    def test_ignore_file(self):
        self.add_file(".abstraignore", "ignored")
        self.add_file("ignored", "foo")
        tracked = self.add_file("tracked", "bar")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [tracked, *self.initial_files])

    def test_ignore_file_pattern(self):
        self.add_file(".abstraignore", "~$*.xlsx")
        self.add_file("~$file.xlsx", "foo")
        self.add_file("~$planilha.xlsx", "foo")
        tracked = self.add_file("tracked", "bar")

        files = files_from_directory(self.path)
        self.assertSamePaths(files, [tracked, *self.initial_files])

    def test_ignore_empty(self):
        self.add_folder("ignored")
        self.add_folder("empty")
        self.add_file(".abstraignore", "ignored")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [*self.initial_files])

    def test_ignore_folder(self):
        ignored = self.add_folder("ignored")
        self.add_file("abc", "foo", path=ignored)
        ignored2 = self.add_folder("ignored2")
        self.add_file("xyz", "foo", path=ignored2)
        folder = self.add_folder("tracked")
        tracked = self.add_file("tracked", "tracked", path=folder)
        self.add_file(".abstraignore", "ignored\nignored2/")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [tracked, *self.initial_files])

    def test_ignore_wildcard(self):
        self.add_file("ignored.ipynb", "foo")
        tracked = self.add_file("tracked", "bar")
        self.add_file(".abstraignore", "*.ipynb")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [tracked, *self.initial_files])

    def test_ignore_abstra_folder(self):
        abstra = self.add_folder(DOT_ABSTRA_FOLDER_NAME)
        self.add_file("credentials", "SECRET", path=abstra)
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [*self.initial_files])

    def test_ignore_git_folder(self):
        git = self.add_folder(".git")
        self.add_file("somefile", "somecontent", path=git)
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [*self.initial_files])

    def test_ignore_abstra_ignore_file(self):
        self.add_file(".abstraignore", "ignored")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [*self.initial_files])

    def test_ignore_git_ignore_file(self):
        self.add_file(".gitignore", "ignored")
        files = files_from_directory(self.path)
        self.assertSamePaths(files, [*self.initial_files])


class CreatePathTest(unittest.TestCase):
    def test_without_conflict(self):
        self.assertEqual(generate_conflictless_path("file"), "file")

    def test_conflict_with_static_path(self):
        path = "login"
        generated_path = generate_conflictless_path(path)

        self.assertNotEqual(generated_path, path)
        self.assertIn(path + "-", generated_path)

    def test_conflict_with_dynamic_path(self):
        path = "error/some-error"
        generated_path = generate_conflictless_path(path)

        self.assertNotEqual(generated_path, path)
        self.assertIn(path + "-", generated_path)



import flask
import flask_sock

from abstra_internals.constants import STAGE_RUN_ID_PARAM_KEY
from abstra_internals.controllers.execution import (
    ExecutionController,
    LockFailedException,
    UnsetThreadException,
)
from abstra_internals.controllers.execution_client_form import FormClient
from abstra_internals.controllers.main import MainController
from abstra_internals.entities.execution import context_from_flask
from abstra_internals.logger import AbstraLogger
from abstra_internals.usage import editor_usage
from abstra_internals.utils import is_it_true


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_forms", __name__)
    sock = flask_sock.Sock(bp)

    @sock.route("/socket")
    def _websocket(ws: flask_sock.Server):
        try:
            request_context = context_from_flask(flask.request)

            id = flask.request.args.get("id")
            if id is None:
                return

            form = controller.get_form(id)
            if not form:
                return

            is_detached = flask.request.args.get(
                "detached", default=False, type=is_it_true
            )

            client = FormClient(
                ws=ws, request_context=request_context, production_mode=False
            )

            if is_detached:
                ExecutionController(
                    workflow_engine=controller.detached_workflow_engine,
                    execution_repository=controller.execution_repository,
                    stage_run_repository=controller.stage_run_repository,
                ).test(
                    stage=form,
                    client=client,
                    request=request_context,
                )
            else:
                ExecutionController(
                    workflow_engine=controller.workflow_engine,
                    execution_repository=controller.execution_repository,
                    stage_run_repository=controller.stage_run_repository,
                ).run(
                    stage=form,
                    client=client,
                    request=request_context,
                    target_stage_run_id=flask.request.args.get(STAGE_RUN_ID_PARAM_KEY),
                )
        except LockFailedException:
            pass
        except UnsetThreadException:
            pass
        except Exception as e:
            AbstraLogger.capture_exception(e)
        finally:
            ws.close(message="Done")

    @bp.get("/")
    @editor_usage
    def _get_forms():
        return [f.editor_dto for f in controller.get_forms()]

    @bp.get("/<path:id>")
    @editor_usage
    def _get_form(id: str):
        form = controller.get_form(id)
        if not form:
            flask.abort(404)
        return form.editor_dto

    @bp.delete("/<path:id>")
    @editor_usage
    def _delete_form(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_form(id, remove_file)
        return {"success": True}

    @bp.post("/")
    @editor_usage
    def _create_form():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        workflow_position = data.get("position", (0, 0))
        form = controller.create_form(title, file, workflow_position)
        return form.editor_dto

    @bp.put("/<path:id>")
    @editor_usage
    def _update_form(id: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        form = controller.update_stage(id, data)
        return form.editor_dto if form else None

    return bp



import inspect
import mimetypes
import typing
from pathlib import Path

import flask
import werkzeug.exceptions as wz_ex

import abstra_statics

dist_folder = Path(inspect.getfile(abstra_statics)).joinpath("../dist").resolve()


def send_from_dist(
    filename: str,
    fallback: typing.Optional[str] = None,
    dist_folder: Path = dist_folder,
):
    mimetypes.add_type("application/javascript", ".js")
    mimetypes.add_type("text/css", ".css")
    mimetypes.add_type("text/html", ".html")

    try:
        return flask.send_from_directory(dist_folder, filename)
    except wz_ex.NotFound:
        if fallback is None:
            return flask.Response(status=404)
        return flask.send_from_directory(dist_folder, fallback)



import sys

import flask

from abstra_internals.controllers.main import MainController
from abstra_internals.debugger.vscode import (
    configure_launch_json,
    configure_settings_json,
    is_client_connected,
    is_launch_json_configured,
    is_settings_json_set,
    start_debugger,
)
from abstra_internals.settings import Settings
from abstra_internals.usage import editor_usage


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_debugger", __name__)

    @bp.get("/")
    def _get_debugger_status():
        return dict(
            port=start_debugger(),
            connected=is_client_connected(),
            is_launch_json_configured=is_launch_json_configured()
            and is_settings_json_set(),
            interpreter=sys.executable,
            root=str(Settings.root_path),
        )

    @bp.post("/vscode-launch")
    @editor_usage
    def _create_vscode_launch():
        configure_launch_json()
        configure_settings_json()
        return {"status": "ok"}

    return bp



from unittest import TestCase

from abstra_internals.modules import import_as_new, reload_project_local_modules
from abstra_internals.repositories.project.project import ProjectRepository, ScriptStage
from tests.fixtures import clear_dir, init_dir


class TestImportAsNew(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_export_variables(self):
        code = "x = 1"
        file = self.root / "test.py"
        file.write_text(code)
        module = import_as_new("test.py")
        self.assertEqual(module.x, 1)

    def test_reload_variables(self):
        old_code = "x = 1"
        file = self.root / "test.py"
        file.write_text(old_code)
        module = import_as_new("test.py")
        self.assertEqual(module.x, 1)
        new_code = "x = 2"
        file.write_text(new_code)
        module = import_as_new("test.py")
        self.assertEqual(module.x, 2)


class ReloadProjectLocalModulesTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_reload(self):
        project = ProjectRepository.load()
        mod_code = "x = 1"
        mod_file = self.root / "my_mod.py"
        mod_file.write_text(mod_code)

        main_code = "from my_mod import x"
        main_file = self.root / "main.py"
        main_file.write_text(main_code)
        script = ScriptStage(
            file="main.py",
            id="test",
            is_initial=True,
            title="test",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.scripts.append(script)
        ProjectRepository.save(project)

        main_module = import_as_new("main.py")
        self.assertEqual(main_module.x, 1)

        mod_code = "x = 2"
        mod_file.write_text(mod_code)
        reload_project_local_modules()
        main_module = import_as_new("main.py")
        self.assertEqual(main_module.x, 2)



import json

import requests

from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import CLOUD_API_CLI_URL
from abstra_internals.interface.cli.tables_messages import (
    bad_json_error_message,
    dump_message,
    file_not_found_error_message,
    generic_error,
    restore_message,
    zod_error_message,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.settings import Settings
from abstra_internals.utils.file import ABSTRA_TABLES_FILE


def dump():
    headers = resolve_headers()
    url = f"{CLOUD_API_CLI_URL}/tables/dump"
    try:
        res = requests.get(url, headers=headers)
        res.raise_for_status()
        res_json = res.json()
        json_str = json.dumps(res_json, indent=2)
        Settings.root_path.joinpath(ABSTRA_TABLES_FILE).write_text(
            json_str, encoding="utf-8"
        )
        num_tables = len(res_json["tables"])
        dump_message(num_tables)
    except Exception as e:
        AbstraLogger.capture_exception(e)
        generic_error()


def restore():
    headers = resolve_headers()
    url = f"{CLOUD_API_CLI_URL}/tables/restore"
    try:
        json_content = Settings.root_path.joinpath(ABSTRA_TABLES_FILE).read_text(
            encoding="utf-8"
        )
        tables_to_restore = json.loads(json_content)
        res = requests.post(url, json=tables_to_restore, headers=headers)
        if res.status_code == 400:
            zod_error_message()
            return
        res.raise_for_status()
        num_tables = len(tables_to_restore["tables"])
        restore_message(num_tables)
    except FileNotFoundError:
        file_not_found_error_message()
    except json.JSONDecodeError:
        bad_json_error_message()
    except Exception as e:
        AbstraLogger.capture_exception(e)
        generic_error()



import flask

from abstra_internals.controllers.env_vars import PatchClientEnvVarBody
from abstra_internals.controllers.main import MainController
from abstra_internals.usage import editor_usage


def get_editor_bp(main_controller: MainController):
    bp = flask.Blueprint("editor_env_vars", __name__)

    # 1s pooling in this route
    @bp.get("/")
    def _get_env_vars():
        return [
            env_var.to_dict() for env_var in main_controller.env_vars_repository.list()
        ]

    @bp.patch("/")
    @editor_usage
    def _change_env_vars():
        if flask.request.json is None:
            flask.abort(400)

        body = PatchClientEnvVarBody.from_dict(flask.request.json)
        for change in body.changes:
            if change.change == "create" or change.change == "update":
                main_controller.env_vars_repository.set(
                    name=change.name, value=change.value
                )
            elif change.change == "delete":
                main_controller.env_vars_repository.unset(name=change.name)

        return {"status": "ok"}

    return bp



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, TypedDict

import requests

from abstra_internals.environment import SIDECAR_HEADERS, SIDECAR_URL
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.serializer import SerializationHelper
from abstra_internals.threaded import threaded
from abstra_internals.utils import serialize
from abstra_internals.utils.datetime import from_utc_iso_string, to_utc_iso_string
from abstra_internals.utils.dot_abstra import LOCAL_LOGS_FOLDER


@dataclass
class LogEntry:
    execution_id: str
    created_at: datetime
    event: Literal["stderr", "stdout"]
    payload: Dict[Literal["text"], str]
    sequence: int

    def to_dto(self) -> Dict[str, Any]:
        return {
            "executionId": self.execution_id,
            "createdAt": to_utc_iso_string(self.created_at),
            "event": self.event,
            "payload": self.payload,
            "sequence": self.sequence,
        }

    @staticmethod
    def from_dto(dto: Dict) -> "LogEntry":
        if dto.get("event") == "stdout" or dto.get("event") == "stderr":
            return LogEntry(
                execution_id=dto["executionId"],
                created_at=from_utc_iso_string(dto["createdAt"]),
                event=dto["event"],
                payload=dto["payload"],
                sequence=dto["sequence"],
            )
        else:
            raise Exception("Invalid log entry type")


LogEvent = Literal["stdout", "stderr"]
LogsDTO = TypedDict("LogsDTO", {"type": str, "text": str})


class ExecutionLogsRepository(ABC):
    sequence: int

    def get_sequence(self) -> int:
        self.sequence += 1
        return self.sequence

    def insert_stdio(
        self, execution_id: str, event: Literal["stdout", "stderr"], text: str
    ):
        self.save(
            LogEntry(
                execution_id=execution_id,
                created_at=datetime.now(),
                event=event,
                payload={"text": text},
                sequence=self.get_sequence(),
            )
        )

    @abstractmethod
    def save(self, log_entry: LogEntry) -> None:
        raise NotImplementedError()

    @abstractmethod
    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        raise NotImplementedError()

    def get_logs_dto(self, execution_id: str) -> List[LogsDTO]:
        logs = self.get(execution_id)
        output: List[LogsDTO] = []
        for entry in logs:
            text = entry.payload.get("text")
            if not text:
                continue

            if entry.event == "stdout":
                output.append({"type": "stdout", "text": text})
            elif entry.event == "stderr":
                output.append({"type": "stderr", "text": text})
            elif entry.event == "unhandled-exception":
                output.append({"type": "stderr", "text": text})

        return output


class LocalExecutionLogsRepository(ExecutionLogsRepository):
    _logs: Dict[str, List[LogEntry]]

    def __init__(self):
        self.sequence = 0
        self._logs = {}

    def save(self, log_entry: LogEntry) -> None:
        execution_id = log_entry.execution_id

        if not execution_id:
            return

        log_file = Path(LOCAL_LOGS_FOLDER) / f"{execution_id}.log"

        if not log_file.parent.exists():
            log_file.parent.mkdir(parents=True)

        with open(log_file, "a", encoding="utf-8") as f:
            f.write(serialize(log_entry.to_dto()) + "\n")

    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        log_file = Path(LOCAL_LOGS_FOLDER) / f"{execution_id}.log"

        if not log_file.exists():
            return []

        try:
            with open(log_file, "r", encoding="utf-8") as f:
                logs: List[LogEntry] = []
                for line in f.readlines():
                    dto = json.loads(line)
                    if event and dto["event"] != event:
                        continue
                    logs.append(LogEntry.from_dto(dto))
                return logs
        except Exception as e:
            AbstraLogger.capture_exception(e)
            return []


class RemoteExecutionLogsRepository(ExecutionLogsRepository):
    def __init__(self, url: str, headers: Dict[str, str]):
        self.url = url
        self.headers = headers
        self.sequence = 0

    @threaded
    def save(self, log_entry: LogEntry) -> None:
        validated_payload = SerializationHelper.enforce_max_size(log_entry.payload)

        dto = {
            **log_entry.to_dto(),
            "payload": validated_payload,
        }

        res = requests.post(
            f"{self.url}/executions/{log_entry.execution_id}/logs",
            json=dto,
            headers=self.headers,
        )

        res.raise_for_status()

    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        response = requests.get(
            f"{self.url}/executions/{execution_id}/logs",
            params={"event": event} if event else None,
            headers=self.headers,
        )

        response.raise_for_status()

        return [LogEntry.from_dto(log) for log in response.json()]


def execution_logs_repository_factory() -> ExecutionLogsRepository:
    if SIDECAR_URL:
        return RemoteExecutionLogsRepository(url=SIDECAR_URL, headers=SIDECAR_HEADERS)
    else:
        return LocalExecutionLogsRepository()



import abc
from dataclasses import dataclass
from typing import List, Literal

import requests

from abstra_internals.credentials import resolve_headers
from abstra_internals.environment import CLOUD_API_CLI_URL, SIDECAR_HEADERS, SIDECAR_URL

Kind = Literal["passwordless", "thread-waiting", "message"]


@dataclass
class EmailParams:
    kind: Kind
    to: List[str]
    subject: str
    body: str
    is_html: bool

    def to_dict(self):
        return {
            "kind": self.kind,
            "to": self.to,
            "subject": self.subject,
            "body": self.body,
            "isHtml": self.is_html,
        }


class EmailRepository(abc.ABC):
    def send(self, param: EmailParams):
        raise NotImplementedError()


class ProductionEmailRepository(EmailRepository):
    base_url: str

    def __init__(self, sidecar_url: str):
        self.base_url = sidecar_url + "/email"

    def send(self, param: EmailParams):
        requests.post(
            self.base_url, json=param.to_dict(), headers=SIDECAR_HEADERS
        ).raise_for_status()


class LocalEmailRepository(EmailRepository):
    base_url: str

    def __init__(self, cloud_api_cli_url: str):
        self.base_url = cloud_api_cli_url + "/email"

    def send(self, param: EmailParams):
        headers = resolve_headers()
        if headers is None:
            raise Exception("You must be logged in to send an email")
        requests.post(
            self.base_url, json=param.to_dict(), headers=headers
        ).raise_for_status()


class TestEmailRepository(EmailRepository):
    def send(self, param: EmailParams):
        pass


def email_repository_factory() -> EmailRepository:
    if SIDECAR_URL is None:
        return LocalEmailRepository(CLOUD_API_CLI_URL)
    else:
        return ProductionEmailRepository(SIDECAR_URL)



from abstra_internals.contracts_generated import (
    AbstraLibApiEditorLintersCheckResponse,
    AbstraLibApiEditorLintersRule,
)
from abstra_internals.linter.rules import rules


def check_linters() -> AbstraLibApiEditorLintersCheckResponse:
    checks = []

    for rule in rules:
        try:
            checks.append(AbstraLibApiEditorLintersRule.from_dict(rule.to_dict()))
        except Exception:
            pass

    return checks


def fix_linter(rule_name: str, fix_name: str):
    for rule in rules:
        if rule.name == rule_name:
            for issue in rule.find_issues():
                for fix in issue.fixes:
                    if fix.name == fix_name:
                        fix.fix()
                        return True
    raise Exception(f"Could not find fix {fix_name} for rule {rule_name}")


def fix_all_linters():
    for rule in rules:
        if rule.type != "info":
            for issue in rule.find_issues():
                for fix in issue.fixes:
                    fix.fix()



import json
from threading import Thread
from typing import Optional

from abstra_internals.controllers.execution_client import (
    ExecutionClient,
    HeadlessClient,
)
from abstra_internals.controllers.execution_target import ExecutionTarget
from abstra_internals.controllers.workflow_interface import IWorkflowEngine
from abstra_internals.entities.execution import Execution, RequestContext
from abstra_internals.repositories.execution import ExecutionRepository
from abstra_internals.repositories.project.project import (
    ActionStage,
    ProjectRepository,
)
from abstra_internals.repositories.stage_run import StageRunRepository
from abstra_internals.settings import Settings
from abstra_internals.utils.dot_abstra import TEST_DATA_FILE


def read_test_data():
    test_file = Settings.root_path / TEST_DATA_FILE
    if not test_file.is_file():
        return {}

    return json.loads(test_file.read_text(encoding="utf-8"))


class LockFailedException(Exception):
    pass


class UnsetThreadException(Exception):
    pass


class NotStartedException(Exception):
    pass


class ExecutionController:
    def __init__(
        self,
        *,
        workflow_engine: IWorkflowEngine,
        stage_run_repository: StageRunRepository,
        execution_repository: ExecutionRepository,
    ) -> None:
        self.workflow_engine = workflow_engine
        self.stage_run_repository = stage_run_repository
        self.execution_repository = execution_repository

    def run(
        self,
        *,
        wait=True,
        stage: ActionStage,
        client: Optional[ExecutionClient] = None,
        request: Optional[RequestContext] = None,
        target_stage_run_id: Optional[str] = None,
    ):
        if not client:
            client = HeadlessClient()

        if not target_stage_run_id:
            if not ProjectRepository.load().is_initial(stage):
                client.handle_unset_thread()
                raise UnsetThreadException()

            target_stage_run_id = self.stage_run_repository.create_initial(stage.id).id

        target_stage_run_id = self.stage_run_repository.ensure_not_abandoned(
            target_stage_run_id
        )

        execution = Execution.create(
            stage_run_id=target_stage_run_id,
            request_context=request,
            stage_id=stage.id,
        )

        if not self.stage_run_repository.acquire_lock(
            stage_run_id=target_stage_run_id, execution_id=execution.id
        ):
            client.handle_lock_failed(target_stage_run_id)
            raise LockFailedException()

        pthread = Thread(
            target=ExecutionTarget,
            kwargs=dict(
                stage=stage,
                client=client,
                execution=execution,
                workflow_engine=self.workflow_engine,
                stage_run_repository=self.stage_run_repository,
                execution_repository=self.execution_repository,
            ),
            name=f"{stage.title} - {execution.short_id}",
        )

        pthread.start()
        if wait:
            pthread.join()

    def test(
        self,
        stage: ActionStage,
        client: Optional[ExecutionClient] = None,
        request: Optional[RequestContext] = None,
    ):
        detached_stage_run = self.stage_run_repository.create_detached(
            thread_data=read_test_data(),
            stage_id=stage.id,
        )

        return self.run(
            stage=stage,
            client=client,
            request=request,
            target_stage_run_id=detached_stage_run.id,
        )
